// -----------------------------------------------------------------------------
//
// Copyright 2011(c) Analog Devices, Inc.
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//  - Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  - Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  - Neither the name of Analog Devices, Inc. nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//  - The use of this software may or may not infringe the patent rights
//    of one or more patent holders.  This license does not release you
//    from the requirement that you obtain separate licenses from these
//    patent holders to use this software.
//  - Use of the software either in source or binary form, must be run
//    on or directly connected to an Analog Devices Inc. component.
//
// THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// INTELLECTUAL PROPERTY RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// -----------------------------------------------------------------------------
// FILE NAME : AD7328.v
// MODULE NAME : AD7328
// AUTHOR :acostina
// AUTHORS EMAIL : adrian.costina@analog.com
// -----------------------------------------------------------------------------
// SVN REVISION: $WCREV$
// -----------------------------------------------------------------------------
// KEYWORDS : AD7328
// -----------------------------------------------------------------------------
// PURPOSE : Driver for the AD7328 8-Channel, Software-Selectable, True
// Bipolar Input, 12 Bit Plus Sign ADC
// -----------------------------------------------------------------------------
// REUSE ISSUES
// Reset Strategy      : Active low reset signal
// Clock Domains       : fpga_clk_i 100 MHz
//                       adc_clk_i  20  MHz
// Critical Timing     : N/A
// Test Features       : N/A
// Asynchronous I/F    : N/A
// Instantiations      : N/A
// Synthesizable (y/n) : Y
// Target Device       : AD7328
// Other               : The driver is intended to be used with AD7328
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

`timescale 1 ns / 100 ps //Use a timescale that is best for simulation.

//------------------------------------------------------------------------------
//----------- Module Declaration -----------------------------------------------
//------------------------------------------------------------------------------

module AD7328

//----------- Ports Declarations -----------------------------------------------
(
    //clock and reset signals
    input               fpga_clk_i,      // system clock
    input               adc_clk_i,       // clock to be applied to ADC to read the conversions results
    input               reset_n_i,       // active low reset signal
    input       [15:0]  data_i,          // data read from the ADC

    //IP control and data interface
    input               wr_data_n_i,
    output reg  [15:0]  data_o,          // data read from the ADC
    output reg          data_rd_ready_o, // when set to high the data read from the ADC is available on the data_o bus
    output reg          data_wr_ready_o,

    //ADC control and data interface
    input               adc_sdo,        // ADC DOUT pin
    output reg          adc_sdi,        // ADC DIN pin
    output              adc_sclk_o,     // ADC SCLK pin
    output reg          adc_cs_n_o      // ADC \CS pin
);

//------------------------------------------------------------------------------
//----------- Registers Declarations -------------------------------------------
//------------------------------------------------------------------------------
reg [ 4:0]  adc_state;      //current state for the ADC control state machine
reg [ 4:0]  adc_next_state; //next state for the ADC control state machine

reg [15:0]  adc_tcycle_cnt; //counts the number of FPGA clock cycles to determine when an ADC cycle is complete
reg [ 4:0]  sclk_clk_cnt;   //counts the number of clocks applied to the ADC to read the conversion result

reg         adc_clk_en;     //gating signal for the clock sent to the ADC
reg [15:0]  data_i_reg;

//------------------------------------------------------------------------------
//----------- Wires Declarations -----------------------------------------------
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//----------- Local Parameters -------------------------------------------------
//------------------------------------------------------------------------------
// ADC states
parameter ADC_IDLE_STATE            = 5'b00001;
parameter ADC_START_CNV_STATE       = 5'b00010;
parameter ADC_READ_CNV_RESULT       = 5'b00100;
parameter ADC_TRANSFER_RESULT       = 5'b01000;
parameter ADC_QUIET                 = 5'b10000;

// ADC timing
parameter real      FPGA_CLOCK_FREQ  = 100;    // FPGA clock frequency [MHz]
parameter real      ADC_CYCLE_TIME   = 1;  // minimum time between two ADC conversions  [us]
parameter [15:0]    ADC_CYCLE_CNT    = FPGA_CLOCK_FREQ * ADC_CYCLE_TIME - 2;

// ADC serial clock periods
parameter ADC_SCLK_PERIODS  = 5'd17; // number of clocks to be sent to the ADC to read the conversion result

//------------------------------------------------------------------------------
//----------- Assign/Always Blocks ---------------------------------------------
//------------------------------------------------------------------------------
assign adc_sclk_o     =   adc_clk_en ? adc_clk_i: 1'b1 ;

//update the ADC timing counters
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        adc_tcycle_cnt  <= ADC_CYCLE_CNT;
    end
    else
    begin
        if(adc_state == ADC_IDLE_STATE)
        begin
            adc_tcycle_cnt <= ADC_CYCLE_CNT ;
        end
        else if(adc_tcycle_cnt != 0)
        begin
            adc_tcycle_cnt <= adc_tcycle_cnt - 16'h1;
        end
    end
end

// Read/write data from/to the ADC
always @(negedge adc_clk_i)
begin
    if(adc_clk_en == 1'b1)
    begin
        data_o          <= {data_o[14:0], adc_sdo};
        sclk_clk_cnt    <= sclk_clk_cnt - 5'h1;
        adc_sdi         <= data_i_reg [15];
        data_i_reg      <= {data_i_reg[14:0],1'b0};
        data_wr_ready_o <= 1'b0;
    end
    else
    begin
        sclk_clk_cnt    <= ADC_SCLK_PERIODS ;
        adc_sdi         <= data_i[15];
        if(wr_data_n_i == 1'b0)
        begin
            data_i_reg      <= {data_i[14:0],1'b0};
            data_wr_ready_o   <= 1'b1;
        end
        else
        begin
            data_wr_ready_o   <= 1'b0;
        end
    end
end

// Determine when the ADC clock is valid to be sent to the ADC
// Used adc_clk_i for synchronization with the ADC clock
always @(posedge adc_clk_i)
begin
    adc_clk_en   <= (adc_state == ADC_READ_CNV_RESULT)  ;
end

//update the ADC current state and the control signals
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        adc_state <= ADC_IDLE_STATE;
    end
    else
    begin
        adc_state <= adc_next_state;
        case (adc_state)
            ADC_IDLE_STATE:
            begin
                adc_cs_n_o      <= 1'b1;
                data_rd_ready_o <= 1'b0;
            end
            ADC_START_CNV_STATE:
            begin
                adc_cs_n_o      <= 1'b0;
                data_rd_ready_o <= 1'b0;
            end
            ADC_READ_CNV_RESULT:
            begin
                adc_cs_n_o      <= 1'b0;
                data_rd_ready_o <= 1'b0;
            end
            ADC_TRANSFER_RESULT:
            begin
                adc_cs_n_o      <= adc_clk_en ? 1'b0 : 1'b1;
                data_rd_ready_o <= 1'b1;
            end
            ADC_QUIET:
            begin
                adc_cs_n_o      <= 1'b1;
                data_rd_ready_o <= 1'b0;
            end
        endcase
    end
end

//update the ADC next state
always @(adc_state, adc_tcycle_cnt,  sclk_clk_cnt)
begin
    adc_next_state <= adc_state;
    case (adc_state)
        ADC_IDLE_STATE:
        begin
            adc_next_state <= ADC_START_CNV_STATE;
        end
        ADC_START_CNV_STATE:
        begin
                adc_next_state <= ADC_READ_CNV_RESULT;
        end
        ADC_READ_CNV_RESULT:
        begin
            if(sclk_clk_cnt == 1)
            begin
                adc_next_state <= ADC_TRANSFER_RESULT;
            end
        end
        ADC_TRANSFER_RESULT:
        begin
             adc_next_state <= ADC_QUIET;
         end
        ADC_QUIET:
        begin
            if(adc_tcycle_cnt == 0)
            begin
                adc_next_state <= ADC_IDLE_STATE;
            end
        end
        default:
        begin
            adc_next_state <= ADC_IDLE_STATE;
        end
    endcase
end

endmodule
